function [yk,k,enoise,delta,Y,rho,lambda,Z] = cose(A,L,b,opts)
%COSE	regularization by comparison of solution estimator
%   y=cose(A,L,b) computes the regularized solution of the least squares
%   problem min||L*x||, subject to min||A*x-b||, where L is a regularization
%   matrix. For a standard form problem, L may be set to eye(n) or,
%   equivalently, to 1.
%
%   To solve the problem min||L*(x-x0)||, subject to min||A*x-b||, the user
%   must set opts.x0 to the actual value of x0 and call y=cose(A,L,b,opts).
%
%   [y k enoise delta Y rho lambda Z]=cose(...) returns also the TGSVD
%   regularization parameter k, the estimated noise level enoise, the
%   quantities delta being minimized, the T(G)SVD solutions Y, the
%   corresponding residuals rho, the Tikhonov parameters lambda and solutions Z.
%
%   [...]=cose(A,L,b,opts) allows to specify various options:
%   opts.absrel='abs','rel' switches between minimization of absolute or
%   relative errors (def. 'abs'); opts.show sets the number of the figure where
%   to plot intermediate results (def. no plot), opts.Y,rho,U,d,VZ may be
%   passed in order to save time between successive calls, otherwise they are
%   computed.

%   M. E. Hochstenbach, L. Reichel, and G. Rodriguez. 
%   Regularization parameter determination for discrete ill-posed problems. 
%   J. Comput. Appl. Math., 273:132-149, 2015. 
%   https://doi.org/10.1016/j.cam.2014.06.004
%
%   Y. Park, L. Reichel, G. Rodriguez, and X. Yu. 
%   Parameter determination for Tikhonov regularization problems in general
%   form. J. Comput. Appl. Math., 2018. 
%   https://doi.org/10.1016/j.cam.2018.04.049

if nargin < 4 || isempty(opts), opts = struct('empty','empty'); end
if nargin < 3, error('Too few input arguments.'); end

[mL,nL]=size(L);
noL = isequal(L,1) || ((mL==nL) && isequal(L,speye(mL)));

nam = fieldnames(opts);
% set the x0 vector
if ~any(strcmpi('x0',nam)), opts.x0 = []; end
% minimization of absolute or relative errors
if ~any(strcmpi('absrel',nam)), opts.absrel = 'absolute'; end
% show intermediate results
if ~any(strcmpi('show',nam)), opts.show = 0; end
% T(G)SVD solution (previously computed)
if ~any(strcmpi('Y',nam)), opts.Y = []; end
% T(G)SVD residuals (previously computed)
if ~any(strcmpi('rho',nam)), opts.rho = []; end
% (G)SVD decomposition (previously computed)
if ~any(strcmpi('U',nam)), opts.U = []; end
% (G)SVD decomposition (previously computed)
if ~any(strcmpi('d',nam)), opts.d = []; end
% (G)SVD decomposition (previously computed)
if ~any(strcmpi('VZ',nam)), opts.VZ = []; end

if ~isempty(opts.x0)
	if ~all(size(opts.x0)==[size(A,2),1])
		error('X0 option must be a column vector of lenght n.')
	end
	b = b-A*opts.x0;
end

Y = opts.Y; rho = opts.rho; U = opts.U; d = opts.d; VZ = opts.VZ;
if isempty(Y) || isempty(rho) || isempty(U) || isempty(d) || isempty(VZ)
	if noL
		[U d VZ] = csvd(A);
		[Y rho eta] = tsvd(U,d,VZ,b,[1:size(d,1)]');
	else
		[U d VZ VV] = cgsvd(A,L);
		[Y rho eta] = tgsvd(U,d,VZ,b,[1:size(d,1)]');
	end
end

if isempty(opts.absrel) || (lower(opts.absrel(1))=='a')
	relative = 0;
elseif (lower(opts.absrel(1))=='r')
	relative = 1;
else
	error('ABSREL must be set to either ''relative'' or ''absolute''.')
end

p = size(d,1);
n = size(Y,1);
q = size(Y,2);
delta = zeros(q,1);

if any(size(A) ~= size(U))
	error('The compact SVD of A is required.')
end

if nargout > 7
	keepz = 1;
	Z = zeros(n,q);
else
	keepz = 0;
end

% reduce number of points: delete 10% of them
nn = round(.9*q);
lambda = zeros(nn,1);

c = U'*b;
bproj = norm(b - U*c)^2;
c2 = c(1:p).^2;
s2 = d(:,1).^2;
if size(d,2) > 1
	mu2 = d(:,2).^2;
	mu4 = d(:,2).^4;
	smu = s2.*mu4;
else
	mu2 = 1;
	mu4 = 1;
	smu = s2;
end
tau0 = 1e-50;
for i = 1:nn
	rhobar = abs(rho(i)^2-bproj);
	fun = @(tau) tau + .5*(c2'*(mu4./(s2*tau+mu2).^2)-rhobar) / ...
				(c2'*(smu./(s2*tau+mu2).^3));
	[tau nit] = fixedpoint( fun, tau0);
	if tau<0, warning('cose:negtau','Negative tau.'), end
	lambda(i) = 1/sqrt(tau);
	xlambda = tikhonov(U,d,VZ,b,lambda(i));
	delta(i) = norm(Y(:,i)-xlambda);
	if relative,  delta(i) = delta(i)/norm(Y(:,i));  end
	if keepz,  Z(:,i) = xlambda;  end
	%	old method, only for L=I
	%[mu xlambda1 kit flagtd] = tikhonov_discrepancy(U,d,VZ,[],b,rho(i));
	%if mu<0, warning('cose:negmu','Negative mu.'), end
	%	show some results for debugging
	%disp(' '), disp(' '), disp(' '), disp(' ')
	%residual = [rho(i)-norm(A*xlambda-b)]
	%parameter = [1/sqrt(tau)]
	%iterations = nit, %pause
	%
	if opts.show>=100
		figure(opts.show+1)
		plot([Y(:,i) xlambda])
		legend('TSVD','Tikhonov')
		title(sprintf('k=%d - \\lambda=%.2e',i,lambda(i)))
		drawnow
		pause(0.1)
	end
end

% find global minimum
delta = delta(1:nn);
[mm k] = min(delta);
if k <= 2
	[m2 k2] = min(delta(k+1:nn));
	k2 = k2+k;
	if k2~=k+1, k = k2; end
end
enoise = rho(k)/norm(b);

if ~isempty(opts.x0)
	Y = Y + opts.x0*ones(1,size(d,1));
	if keepz, Z = Z + opts.x0*ones(1,size(d,1)); end
end
yk = Y(:,k);

if opts.show
	figure(opts.show)
	hh = semilogy([1:size(delta,1)]',delta,k,delta(k),'.r');
	set(hh(2),'markersize',20)
	title(sprintf('delta(k) - minimum at %d',k))
end



function [x,n] = fixedpoint( fun, x0, tol, nmax)

if nargin < 4,  nmax = 100;  end
if nargin < 3,  tol = 1e-4;  end

n = 0;
x = x0;
flag = 1;
while flag
	n = n+1;
	x0 = x;
	x = fun(x);
	flag = (abs(x-x0) > tol*abs(x)) & (n<=nmax);
end
if n>nmax,  warning('cose:maxit','Max number of iterations exceeded.'),  end

