function [yk,ki,enoise,delta,Y,res,lambda,Z,mvp] = cosensb(A,L,W,b,opts)
%COSENSB COSE regularization method in general form via null space basis
%   y=cosensb(A,L,W,b,kmax) computes the regularized solution of the least
%   squares problem min||L*x||, subject to min||A*x-b||, where L is the
%   regularization matrix and the columns of W are a basis for the null space
%   of L. For a standard form problem, L may be set to eye(n). The columns of W
%   are orthonormalized inside the function.
%
%   To solve the problem min||L*(x-x0)||, subject to min||A*x-b||, the user
%   must set opts.x0 to the actual value of x0 and call y=cosegkgt(A,L,b,opts).
%
%   [y,k,enoise,delta,Y,rho,lambda,Z,mvp]=cosensb(...) returns also the
%   regularization parameter k, the estimated noise level enoise, the vector
%   delta being minimized, the solutions Y, the corresponding residuals rho,
%   the Tikhonov parameters lambda and solutions Z, the number of matrix-vector
%   products performed mvp.
%
%   [...]=cosensb(A,L,W,b,opts) allows to specify other informations:
%   opts.absrel='abs','rel' switches between minimization of absolute or
%   relative errors (def. 'abs'), opts.show sets the number of the figure where
%   to plot intermediate results (def. no plot), opts.nsteps is the initial
%   number of GK steps performed, opts.nmax the maximum number of steps,
%   opts.reorth denotes the kind of reorthogonalization (0 no reorth, 1 MGS
%   (def.), 2 CGS, 3 double CGS), opts.bdtol is the tolerance to detect the
%   breakdown of Golub-Kahan iteration.

%   M. E. Hochstenbach, L. Reichel, and G. Rodriguez. 
%   Regularization parameter determination for discrete ill-posed problems. 
%   J. Comput. Appl. Math., 273:132-149, 2015. 
%   https://doi.org/10.1016/j.cam.2014.06.004
%
%   Y. Park, L. Reichel, G. Rodriguez, and X. Yu. 
%   Parameter determination for Tikhonov regularization problems in general
%   form. J. Comput. Appl. Math., 2018. 
%   https://doi.org/10.1016/j.cam.2018.04.049

if nargin < 5 || isempty(opts), opts = struct('empty','empty'); end
if nargin < 4, error('Too few input arguments.'); end

nam = fieldnames(opts);
% set the x0 vector
if ~any(strcmpi('x0',nam)), opts.x0 = []; end
% minimization of absolute or relative errors
if ~any(strcmpi('absrel',nam)), opts.absrel = 'absolute'; end
% show intermediate results
if ~any(strcmpi('show',nam)), opts.show = 0; end
% initial number of GK steps to perform
if ~any(strcmpi('nsteps',nam)), opts.nsteps = 20; end
% max number of GK steps to perform
if ~any(strcmpi('nmax',nam)), opts.nmax = 200; end
% reorthogonalization switch
if ~any(strcmpi('reorth',nam)), opts.reorth = 1; end
% breakdown tolerance
if ~any(strcmpi('bdtol',nam)), opts.bdtol = 1e-12; end

if isempty(opts.absrel) || (lower(opts.absrel(1))=='a')
	relative = 0;
elseif (lower(opts.absrel(1))=='r')
	relative = 1;
else
	error('ABSREL must be set to either ''relative'' or ''absolute''.')
end

[m,n] = size(A);
m1 = size(b,1);
if m ~= m1
	error('Size of A and b must agree.')
end
[pL n1] = size(L);
if n ~= n1
	error('Size of A and L must agree.')
end

if ~isempty(opts.x0)
	if ~all(size(opts.x0)==[n,1])
		error('X0 option must be a column vector of lenght n.')
	end
	b = b-A*opts.x0;
end

mvp = 0;

[W RR] = qr(W,0);	% orthonormalize columns of W
[Q R] = qr(A*W,0);
PA = A - Q*(Q'*A);
mvp = mvp + size(Q,2);
Pb = b - Q*(Q'*b);

[xb,ki,enoise,delta,Y,res,lambda,Z,mvp1] = cosegkgt(PA,L,Pb,opts);
mvp = mvp + mvp1;

y = R \ (Q'*(b - A*xb));
mvp = mvp + 1;
z = W*y;

if ~isempty(opts.x0)
	Y = Y + opts.x0*ones(1,size(Y,2));
	Z = Z + opts.x0*ones(1,size(Z,2));
end
yk = Y(:,ki);

yk = yk + z;
Y = Y + z*ones(1,size(Y,2));
Z = Z + z*ones(1,size(Z,2));

